#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import datetime
from http import HTTPStatus
from unittest import mock

import ddt
from oslo_config import cfg
import oslo_messaging as messaging
from oslo_serialization import jsonutils
import webob

from cinder.api.contrib import volume_actions
from cinder.api import microversions as mv
from cinder.api.openstack import api_version_request as api_version
from cinder import context
from cinder import db
from cinder import exception
from cinder.image import glance
from cinder import objects
from cinder.tests.unit.api import fakes
from cinder.tests.unit.api.v3 import fakes as v3_fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.tests.unit import utils
from cinder import volume
from cinder.volume import api as volume_api
from cinder.volume import rpcapi as volume_rpcapi


CONF = cfg.CONF
ENCRYPTED_VOLUME_ID = 'f78e8977-6164-4114-a593-358fa6646eff'


@ddt.ddt
class VolumeActionsTest(test.TestCase):

    _actions = ('os-reserve', 'os-unreserve')

    _methods = ('attach', 'detach', 'reserve_volume', 'unreserve_volume')

    def setUp(self):
        super(VolumeActionsTest, self).setUp()
        self.context = context.RequestContext(fake.USER_ID, fake.PROJECT_ID,
                                              is_admin=False)
        self.controller = volume_actions.VolumeActionsController()
        self.api_patchers = {}
        for _meth in self._methods:
            self.api_patchers[_meth] = mock.patch('cinder.volume.api.API.' +
                                                  _meth)
            self.api_patchers[_meth].start()
            self.addCleanup(self.api_patchers[_meth].stop)
            self.api_patchers[_meth].return_value = True

        db_vol = {'id': fake.VOLUME_ID, 'host': 'fake', 'status': 'available',
                  'size': 1, 'migration_status': None,
                  'volume_type_id': fake.VOLUME_TYPE_ID,
                  'project_id': fake.PROJECT_ID}
        vol = fake_volume.fake_volume_obj(self.context, **db_vol)
        self.get_patcher = mock.patch('cinder.volume.api.API.get')
        self.mock_volume_get = self.get_patcher.start()
        self.addCleanup(self.get_patcher.stop)
        self.mock_volume_get.return_value = vol
        self.update_patcher = mock.patch('cinder.volume.api.API.update')
        self.mock_volume_update = self.update_patcher.start()
        self.addCleanup(self.update_patcher.stop)
        self.mock_volume_update.return_value = vol
        self.db_get_patcher = mock.patch(
            'cinder.db.sqlalchemy.api._volume_get')
        self.mock_volume_db_get = self.db_get_patcher.start()
        self.addCleanup(self.db_get_patcher.stop)
        self.mock_volume_db_get.return_value = vol

        self.flags(transport_url='fake:/')

    def test_simple_api_actions(self):
        app = fakes.wsgi_app(fake_auth_context=self.context)
        for _action in self._actions:
            req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
            req.method = 'POST'
            req.body = jsonutils.dump_as_bytes({_action: None})
            req.content_type = 'application/json'
            res = req.get_response(app)
            self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_initialize_connection(self):
        with mock.patch.object(volume_api.API,
                               'initialize_connection') as init_conn:
            init_conn.return_value = {}
            body = {'os-initialize_connection': {'connector': {
                'fake': 'fake'}}}
            req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
            req.method = "POST"
            req.body = jsonutils.dump_as_bytes(body)
            req.headers["content-type"] = "application/json"

            res = req.get_response(fakes.wsgi_app(
                fake_auth_context=self.context))
            self.assertEqual(HTTPStatus.OK, res.status_int)

    def test_initialize_connection_without_connector(self):
        with mock.patch.object(volume_api.API,
                               'initialize_connection') as init_conn:
            init_conn.return_value = {}
            body = {'os-initialize_connection': {}}
            req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
            req.method = "POST"
            req.body = jsonutils.dump_as_bytes(body)
            req.headers["content-type"] = "application/json"

            res = req.get_response(fakes.wsgi_app(
                fake_auth_context=self.context))
            self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.initialize_connection')
    def test_initialize_connection_without_initiator(self,
                                                     _init_connection):
        _init_connection.side_effect = messaging.RemoteError('InvalidInput')
        body = {'os-initialize_connection': {'connector': 'w/o_initiator'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
                               fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    def test_initialize_connection_exception(self):
        with mock.patch.object(volume_api.API,
                               'initialize_connection') as init_conn:
            init_conn.side_effect = \
                exception.VolumeBackendAPIException(data=None)
            body = {'os-initialize_connection': {'connector': {
                'fake': 'fake'}}}
            req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
            req.method = "POST"
            req.body = jsonutils.dump_as_bytes(body)
            req.headers["content-type"] = "application/json"

            res = req.get_response(fakes.wsgi_app(
                fake_auth_context=self.context))
            self.assertEqual(HTTPStatus.INTERNAL_SERVER_ERROR,
                             res.status_int)

    def test_terminate_connection(self):
        with mock.patch.object(volume_api.API,
                               'terminate_connection') as terminate_conn:
            terminate_conn.return_value = {}
            body = {'os-terminate_connection': {'connector': 'fake'}}
            req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
            req.method = "POST"
            req.body = jsonutils.dump_as_bytes(body)
            req.headers["content-type"] = "application/json"

            res = req.get_response(fakes.wsgi_app(
                fake_auth_context=self.context))
            self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_terminate_connection_without_connector(self):
        with mock.patch.object(volume_api.API,
                               'terminate_connection') as terminate_conn:
            terminate_conn.return_value = {}
            body = {'os-terminate_connection': {}}
            req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
            req.method = "POST"
            req.body = jsonutils.dump_as_bytes(body)
            req.headers["content-type"] = "application/json"

            res = req.get_response(fakes.wsgi_app(
                fake_auth_context=self.context))
            self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    def test_terminate_connection_with_exception(self):
        with mock.patch.object(volume_api.API,
                               'terminate_connection') as terminate_conn:
            terminate_conn.side_effect = \
                exception.VolumeBackendAPIException(data=None)
            body = {'os-terminate_connection': {'connector': 'fake'}}
            req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
            req.method = "POST"
            req.body = jsonutils.dump_as_bytes(body)
            req.headers["content-type"] = "application/json"

            res = req.get_response(fakes.wsgi_app(
                fake_auth_context=self.context))
            self.assertEqual(HTTPStatus.INTERNAL_SERVER_ERROR,
                             res.status_int)

    def test_attach_to_instance(self):
        body = {'os-attach': {'instance_uuid': fake.INSTANCE_ID,
                              'mountpoint': '/dev/vdc',
                              'mode': 'rw'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

        body = {'os-attach': {'instance_uuid': fake.INSTANCE_ID,
                              'host_name': 'fake_host',
                              'mountpoint': '/dev/vdc'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.headers["content-type"] = "application/json"
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_attach_to_host(self):
        # using 'read-write' mode attach volume by default
        body = {'os-attach': {'host_name': 'fake_host',
                              'mountpoint': '/dev/vdc'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_volume_attach_to_instance_raises_remote_error(self):
        volume_remote_error = \
            messaging.RemoteError(exc_type='InvalidUUID')
        with mock.patch.object(volume_api.API, 'attach',
                               side_effect=volume_remote_error):
            id = fake.VOLUME_ID
            vol = {"instance_uuid": fake.INSTANCE_ID,
                   "mountpoint": "/dev/vdc",
                   "mode": "rw"}
            body = {"os-attach": vol}
            req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                          (fake.PROJECT_ID, id))
            self.assertRaises(webob.exc.HTTPBadRequest,
                              self.controller._attach,
                              req,
                              id,
                              body=body)

    def test_volume_attach_to_instance_raises_db_error(self):
        # In case of DB error 500 error code is returned to user
        volume_remote_error = \
            messaging.RemoteError(exc_type='DBError')
        with mock.patch.object(volume_api.API, 'attach',
                               side_effect=volume_remote_error):
            id = fake.VOLUME_ID
            vol = {"instance_uuid": fake.INSTANCE_ID,
                   "mountpoint": "/dev/vdc",
                   "mode": "rw"}
            body = {"os-attach": vol}
            req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                          (fake.PROJECT_ID, id))
            self.assertRaises(messaging.RemoteError,
                              self.controller._attach,
                              req,
                              id,
                              body=body)

    def test_detach(self):
        body = {'os-detach': {'attachment_id': fake.ATTACHMENT_ID}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_detach_null_attachment_id(self):
        body = {'os-detach': {'attachment_id': None}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_volume_detach_raises_remote_error(self):
        volume_remote_error = \
            messaging.RemoteError(exc_type='VolumeAttachmentNotFound')
        with mock.patch.object(volume_api.API, 'detach',
                               side_effect=volume_remote_error):
            id = fake.VOLUME_ID
            vol = {"attachment_id": fake.ATTACHMENT_ID}
            body = {"os-detach": vol}
            req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                          (fake.PROJECT_ID, id))
            self.assertRaises(webob.exc.HTTPBadRequest,
                              self.controller._detach,
                              req,
                              id,
                              body=body)

    def test_volume_detach_raises_db_error(self):
        # In case of DB error 500 error code is returned to user
        volume_remote_error = \
            messaging.RemoteError(exc_type='DBError')
        with mock.patch.object(volume_api.API, 'detach',
                               side_effect=volume_remote_error):
            id = fake.VOLUME_ID
            vol = {"attachment_id": fake.ATTACHMENT_ID}
            body = {"os-detach": vol}
            req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                          (fake.PROJECT_ID, id))
            self.assertRaises(messaging.RemoteError,
                              self.controller._detach,
                              req,
                              id,
                              body=body)

    def test_attach_with_invalid_arguments(self):
        # Invalid request to attach volume an invalid target
        body = {'os-attach': {'mountpoint': '/dev/vdc'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.headers["content-type"] = "application/json"
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

        # Invalid request to attach volume with an invalid mode
        body = {'os-attach': {'instance_uuid': 'fake',
                              'mountpoint': '/dev/vdc',
                              'mode': 'rr'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.headers["content-type"] = "application/json"
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)
        body = {'os-attach': {'host_name': 'fake_host',
                              'mountpoint': '/dev/vdc',
                              'mode': 'ww'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.headers["content-type"] = "application/json"
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    def test_attach_to_instance_no_mountpoint(self):
        # The mountpoint parameter is required. If not provided the
        # API should fail with a 400 error.
        body = {'os-attach': {'instance_uuid': fake.INSTANCE_ID,
                              'mode': 'rw'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(400, res.status_int)

    def test_begin_detaching(self):
        def fake_begin_detaching(*args, **kwargs):
            return {}
        self.mock_object(volume.api.API, 'begin_detaching',
                         fake_begin_detaching)

        body = {'os-begin_detaching': {'fake': 'fake'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_roll_detaching(self):
        def fake_roll_detaching(*args, **kwargs):
            return {}
        self.mock_object(volume.api.API, 'roll_detaching',
                         fake_roll_detaching)

        body = {'os-roll_detaching': {'fake': 'fake'}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_extend_volume(self):
        def fake_extend_volume(*args, **kwargs):
            return {}
        self.mock_object(volume.api.API, 'extend',
                         fake_extend_volume)

        body = {'os-extend': {'new_size': 5}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)

    def test_extend_volume_invalid_status(self):
        def fake_extend_volume(*args, **kwargs):
            msg = "Volume status must be available"
            raise exception.InvalidVolume(reason=msg)
        self.mock_object(volume.api.API, 'extend',
                         fake_extend_volume)

        body = {'os-extend': {'new_size': 5}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    @ddt.data((True, HTTPStatus.ACCEPTED), (False, HTTPStatus.ACCEPTED),
              ('1', HTTPStatus.ACCEPTED), ('0', HTTPStatus.ACCEPTED),
              ('true', HTTPStatus.ACCEPTED), ('false', HTTPStatus.ACCEPTED),
              ('tt', HTTPStatus.BAD_REQUEST), (11, HTTPStatus.BAD_REQUEST),
              (None, HTTPStatus.BAD_REQUEST))
    @ddt.unpack
    def test_update_readonly_flag(self, readonly, return_code):
        def fake_update_readonly_flag(*args, **kwargs):
            return {}
        self.mock_object(volume.api.API, 'update_readonly_flag',
                         fake_update_readonly_flag)

        body = {"os-update_readonly_flag": {"readonly": readonly}}
        if readonly is None:
            body = {"os-update_readonly_flag": {}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(return_code, res.status_int)

    @ddt.data((True, HTTPStatus.OK), (False, HTTPStatus.OK),
              ('1', HTTPStatus.OK), ('0', HTTPStatus.OK),
              ('true', HTTPStatus.OK), ('false', HTTPStatus.OK),
              ('tt', HTTPStatus.BAD_REQUEST), (11, HTTPStatus.BAD_REQUEST),
              (None, HTTPStatus.BAD_REQUEST))
    @ddt.unpack
    def test_set_bootable(self, bootable, return_code):
        body = {"os-set_bootable": {"bootable": bootable}}
        if bootable is None:
            body = {"os-set_bootable": {}}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, fake.VOLUME_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.context))
        self.assertEqual(return_code, res.status_int)


@ddt.ddt
class VolumeRetypeActionsTest(test.TestCase):
    def setUp(self):
        super(VolumeRetypeActionsTest, self).setUp()

        self.context = context.RequestContext(fake.USER_ID, fake.PROJECT_ID,
                                              is_admin=False)
        self.flags(transport_url='fake:/')

        self.retype_mocks = {}
        paths = ('cinder.quota.QUOTAS.add_volume_type_opts',
                 'cinder.quota.QUOTAS.reserve')
        for path in paths:
            name = path.split('.')[-1]
            patcher = mock.patch(path, return_value=None)
            self.retype_mocks[name] = patcher.start()
            self.addCleanup(patcher.stop)

    @mock.patch('cinder.db.sqlalchemy.api.resource_exists', return_value=True)
    def _retype_volume_exec(self, expected_status,
                            new_type=fake.VOLUME_TYPE2_ID, vol_id=None,
                            exists_mock=None):
        vol_id = vol_id or fake.VOLUME_ID
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, vol_id))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        retype_body = {'new_type': new_type, 'migration_policy': 'never'}
        req.body = jsonutils.dump_as_bytes({'os-retype': retype_body})
        res = req.get_response(fakes.wsgi_app(fake_auth_context=self.context))
        self.assertEqual(expected_status, res.status_int)

    def test_retype_volume_no_body(self):
        # Request with no body should fail
        vol = utils.create_volume(self.context,
                                  status='available',
                                  testcase_instance=self)
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, vol.id))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes({'os-retype': None})
        res = req.get_response(fakes.wsgi_app(fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    def test_retype_volume_bad_policy(self):
        # Request with invalid migration policy should fail
        vol = utils.create_volume(self.context,
                                  status='available',
                                  testcase_instance=self)
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, vol.id))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        retype_body = {'new_type': 'foo', 'migration_policy': 'invalid'}
        req.body = jsonutils.dump_as_bytes({'os-retype': retype_body})
        res = req.get_response(fakes.wsgi_app(fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    def test_retype_volume_bad_status(self):
        # Should fail if volume does not have proper status
        vol_type_old = utils.create_volume_type(context.get_admin_context(),
                                                self, name='old')
        vol_type_new = utils.create_volume_type(context.get_admin_context(),
                                                self, name='new')
        vol = utils.create_volume(self.context,
                                  status='error',
                                  volume_type_id=vol_type_old.id,
                                  testcase_instance=self)

        self._retype_volume_exec(HTTPStatus.BAD_REQUEST, vol_type_new.id,
                                 vol.id)

    def test_retype_type_no_exist(self):
        # Should fail if new type does not exist
        vol_type_old = utils.create_volume_type(context.get_admin_context(),
                                                self, name='old')
        vol = utils.create_volume(self.context,
                                  status='available',
                                  volume_type_id=vol_type_old.id,
                                  testcase_instance=self)
        self._retype_volume_exec(HTTPStatus.NOT_FOUND, 'fake_vol_type',
                                 vol.id)

    def test_retype_same_type(self):
        # Should fail if new type and old type are the same
        vol_type_old = utils.create_volume_type(context.get_admin_context(),
                                                self, name='old')
        vol = utils.create_volume(self.context,
                                  status='available',
                                  volume_type_id=vol_type_old.id,
                                  testcase_instance=self)
        self._retype_volume_exec(HTTPStatus.BAD_REQUEST, vol_type_old.id,
                                 vol.id)

    def test_retype_over_quota(self):
        # Should fail if going over quota for new type
        vol_type_new = utils.create_volume_type(context.get_admin_context(),
                                                self, name='old')
        vol = utils.create_volume(self.context,
                                  status='available',
                                  testcase_instance=self)

        exc = exception.OverQuota(overs=['gigabytes'],
                                  quotas={'gigabytes': 20},
                                  usages={'gigabytes': {'reserved': 5,
                                                        'in_use': 15}})
        self.retype_mocks['reserve'].side_effect = exc
        self._retype_volume_exec(HTTPStatus.REQUEST_ENTITY_TOO_LARGE,
                                 vol_type_new.id, vol.id)

    @ddt.data(('in-use', 'front-end', HTTPStatus.BAD_REQUEST),
              ('in-use', 'back-end', HTTPStatus.ACCEPTED),
              ('available', 'front-end', HTTPStatus.ACCEPTED),
              ('available', 'back-end', HTTPStatus.ACCEPTED),
              ('in-use', 'front-end', HTTPStatus.ACCEPTED, True),
              ('in-use', 'back-end', HTTPStatus.ACCEPTED, True),
              ('available', 'front-end', HTTPStatus.ACCEPTED, True),
              ('available', 'back-end', HTTPStatus.ACCEPTED, True),
              ('in-use', 'front-end', HTTPStatus.BAD_REQUEST, False, False),
              ('in-use', 'back-end', HTTPStatus.ACCEPTED, False, False),
              ('in-use', '', HTTPStatus.ACCEPTED, True, False),
              ('available', 'front-end', HTTPStatus.ACCEPTED, False, False),
              ('available', 'back-end', HTTPStatus.ACCEPTED, False, False),
              ('available', '', HTTPStatus.ACCEPTED, True, False),
              ('in-use', 'front-end', HTTPStatus.BAD_REQUEST, False,
               False, False),
              ('in-use', '', HTTPStatus.ACCEPTED, True, False, False),
              ('in-use', 'back-end', HTTPStatus.ACCEPTED, False,
               False, False),
              ('available', 'front-end', HTTPStatus.ACCEPTED, False,
               False, False),
              ('in-use', '', HTTPStatus.ACCEPTED, True, False, False),
              ('in-use', 'back-end', HTTPStatus.ACCEPTED, False,
               False, False))
    @ddt.unpack
    def test_retype_volume_qos(self, vol_status, consumer_pass,
                               expected_status, same_qos=False, has_qos=True,
                               has_type=True):
        """Test volume retype with QoS

        This test conatins following test-cases:
        1)  should fail if changing qos enforced by front-end for in-use volume
        2)  should NOT fail for in-use if changing qos enforced by back-end
        3)  should NOT fail if changing qos enforced by FE for available
            volumes
        4)  should NOT fail if changing qos enforced by back-end for available
            volumes
        5)  should NOT fail if changing qos enforced by front-end for in-use
            volumes if the qos is the same
        6)  should NOT fail if changing qos enforced by back-end for in-use
            volumes if the qos is the same
        7)  should NOT fail if changing qos enforced by front-end for available
            volumes if the qos is the same
        8)  should NOT fail if changing qos enforced by back-end for available
            volumes if the qos is the same
        9)  should fail if changing qos enforced by front-end on the new type
            and volume originally had no qos and was in-use
        10) should NOT fail if changing qos enforced by back-end on the
            new type and volume originally had no qos and was in-use
        11) should NOT fail if original and destinal types had no qos for
            in-use volumes
        12) should NOT fail if changing qos enforced by front-end on the
            new type and volume originally had no qos and was available
        13) should NOT fail if changing qos enforced by back-end on the
            new type and volume originally had no qos and was available
        14) should NOT fail if original and destinal types had no qos for
            available volumes
        15) should fail if changing volume had no type, was in-use and
            destination type qos was enforced by front-end
        16) should NOT fail if changing volume had no type, was in-use and
            destination type had no qos
            and volume originally had no type and was in-use
        17) should NOT fail if changing volume had no type, was in-use and
            destination type qos was enforced by back-end
        18) should NOT fail if changing volume had no type, was in-use and
            destination type qos was enforced by front-end
        19) should NOT fail if changing volume had no type, was available and
            destination type had no qos
            and volume originally had no type and was in-use
        20) should NOT fail if changing volume had no type, was available and
            destination type qos was enforced by back-end
        """

        admin_ctxt = context.get_admin_context()
        if has_qos:
            qos_old = utils.create_qos(admin_ctxt, self,
                                       name='old',
                                       consumer=consumer_pass)['id']
        else:
            qos_old = None

        if same_qos:
            qos_new = qos_old
        else:
            qos_new = utils.create_qos(admin_ctxt, self,
                                       name='new',
                                       consumer=consumer_pass)['id']

        if has_type:
            vol_type_old = utils.create_volume_type(admin_ctxt, self,
                                                    name='old',
                                                    qos_specs_id=qos_old).id
        else:
            vol_type_old = v3_fakes.fake_default_type_get()['id']

        vol_type_new = utils.create_volume_type(admin_ctxt, self,
                                                name='new',
                                                qos_specs_id=qos_new).id

        vol = utils.create_volume(self.context,
                                  status=vol_status,
                                  volume_type_id=vol_type_old,
                                  testcase_instance=self)

        self._retype_volume_exec(expected_status, vol_type_new, vol.id)

    @ddt.data(('available', HTTPStatus.ACCEPTED, False, False, False),
              ('available', HTTPStatus.ACCEPTED, False, False),
              ('available', HTTPStatus.ACCEPTED, True, False, False),
              ('available', HTTPStatus.ACCEPTED, True, False),
              ('available', HTTPStatus.ACCEPTED))
    @ddt.unpack
    def test_retype_volume_encryption(self, vol_status, expected_status,
                                      has_type=True,
                                      enc_orig=True, enc_dest=True):
        enc_orig = None
        admin_ctxt = context.get_admin_context()
        if has_type:
            vol_type_old = utils.create_volume_type(admin_ctxt, self,
                                                    name='old').id
            if enc_orig:
                utils.create_encryption(admin_ctxt, vol_type_old, self)
        else:
            vol_type_old = v3_fakes.fake_default_type_get()['id']

        vol_type_new = utils.create_volume_type(admin_ctxt, self,
                                                name='new').id
        if enc_dest:
            utils.create_encryption(admin_ctxt, vol_type_new, self)

        vol = utils.create_volume(self.context,
                                  status=vol_status,
                                  volume_type_id=vol_type_old,
                                  testcase_instance=self)

        self._retype_volume_exec(expected_status, vol_type_new, vol.id)


def fake_volume_get(self, context, volume_id):
    volume = v3_fakes.create_volume(volume_id)
    if volume_id == fake.VOLUME3_ID:
        volume['status'] = 'in-use'
    else:
        volume['status'] = 'available'
    return volume


def fake_volume_get_obj(self, context, volume_id, **kwargs):
    volume = fake_volume.fake_volume_obj(context,
                                         id=volume_id,
                                         display_description='displaydesc',
                                         **kwargs)
    if volume_id == fake.VOLUME3_ID:
        volume.status = 'in-use'
    else:
        volume.status = 'available'

    if volume_id == ENCRYPTED_VOLUME_ID:
        volume['encryption_key_id'] = fake.ENCRYPTION_KEY_ID

    volume.volume_type = fake_volume.fake_volume_type_obj(
        context,
        name=v3_fakes.DEFAULT_VOL_TYPE)
    return volume


def fake_upload_volume_to_image_service(self, context, volume, metadata,
                                        force):
    ret = {"id": volume['id'],
           "updated_at": datetime.datetime(1, 1, 1, 1, 1, 1),
           "status": 'uploading',
           "display_description": volume['display_description'],
           "size": volume['size'],
           "volume_type": volume['volume_type'],
           "image_id": fake.IMAGE_ID,
           "container_format": 'bare',
           "disk_format": 'raw',
           "image_name": 'image_name'}
    return ret


@ddt.ddt
class VolumeImageActionsTest(test.TestCase):
    def setUp(self):
        super(VolumeImageActionsTest, self).setUp()
        self.controller = volume_actions.VolumeActionsController()
        self.context = context.RequestContext(fake.USER_ID, fake.PROJECT_ID,
                                              is_admin=False)
        self.maxDiff = 2000

    def _get_os_volume_upload_image(self):
        vol = {
            "container_format": 'bare',
            "disk_format": 'raw',
            "image_name": 'image_name',
            "force": True}
        body = {"os-volume_upload_image": vol}

        return body

    def fake_image_service_create(self, *args):
        ret = {
            'status': 'queued',
            'name': 'image_name',
            'deleted': False,
            'container_format': 'bare',
            'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
            'disk_format': 'raw',
            'updated_at': datetime.datetime(1, 1, 1, 1, 1, 1),
            'id': fake.IMAGE_ID,
            'min_ram': 0,
            'checksum': None,
            'min_disk': 0,
            'deleted_at': None,
            'properties': {'x_billing_code_license': '246254365'},
            'size': 0}
        return ret

    def fake_image_service_create_with_params(self, *args):
        ret = {
            'status': 'queued',
            'name': 'image_name',
            'deleted': False,
            'container_format': 'bare',
            'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
            'disk_format': 'raw',
            'updated_at': datetime.datetime(1, 1, 1, 1, 1, 1),
            'id': fake.IMAGE_ID,
            'min_ram': 0,
            'checksum': None,
            'min_disk': 0,
            'visibility': 'public',
            'protected': True,
            'deleted_at': None,
            'properties': {'x_billing_code_license': '246254365'},
            'size': 0}
        return ret

    def fake_rpc_copy_volume_to_image(self, *args):
        pass

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    @mock.patch.object(volume_api.API, "copy_volume_to_image",
                       fake_upload_volume_to_image_service)
    def test_copy_volume_to_image(self):
        id = fake.VOLUME_ID
        img = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": img}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, id))
        res_dict = self.controller._volume_upload_image(req, id, body=body)
        expected = {'os-volume_upload_image':
                    {'id': id,
                     'updated_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                     'status': 'uploading',
                     'display_description': 'displaydesc',
                     'size': 1,
                     'volume_type': fake_volume.fake_volume_type_obj(
                         context,
                         name='vol_type_name'),
                     'image_id': fake.IMAGE_ID,
                     'container_format': 'bare',
                     'disk_format': 'raw',
                     'image_name': 'image_name'}}
        self.assertDictEqual(expected, res_dict)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    @mock.patch.object(volume_api.API, "copy_volume_to_image")
    def test_copy_volume_to_image_when_image_conversion_not_allowed(
            self,
            mock_copy_vol_to_img):
        """Make sure exception is converted properly."""

        mock_copy_vol_to_img.side_effect = exception.ImageConversionNotAllowed
        id = fake.VOLUME_ID
        img = {"container_format": 'ova',
               "disk_format": 'vhdx',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": img}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, id))

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    @mock.patch.object(volume_api.API, "copy_volume_to_image")
    def test_check_image_metadata_copy_encrypted_volume_to_image(
            self, mock_copy_vol):
        """Make sure the encryption image properties exit the controller."""

        # all we're interested in is that the 'metadata' dict contains the
        # correct data, so we do this bad hack to smuggle it out in the
        # controller's response to make it easy to access
        def really_fake_upload_volume(context, volume, metadata, force):
            return metadata

        mock_copy_vol.side_effect = really_fake_upload_volume

        FAKE_ID = 'fake-encryption-key-id'
        # the controller does a lazy init of the key manager, so we
        # need a 2-level mock here
        self.mock_object(self.controller, '_key_mgr')
        self.controller._key_mgr.return_value = not None
        self.mock_object(self.controller._key_manager, 'store')
        self.controller._key_manager.store.return_value = FAKE_ID

        vol_id = ENCRYPTED_VOLUME_ID
        img = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": img}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, vol_id))
        res_dict = self.controller._volume_upload_image(req, vol_id, body=body)

        sent_meta = res_dict['os-volume_upload_image']
        self.assertIn('cinder_encryption_key_id', sent_meta)
        self.assertEqual(FAKE_ID, sent_meta['cinder_encryption_key_id'])
        self.assertIn('cinder_encryption_key_deletion_policy', sent_meta)
        self.assertEqual('on_image_deletion',
                         sent_meta['cinder_encryption_key_deletion_policy'])

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    @mock.patch.object(volume_api.API, "copy_volume_to_image")
    def test_check_image_metadata_copy_nonencrypted_volume_to_image(
            self, mock_copy_vol):
        """Make sure no encryption image properties are sent."""

        def really_fake_upload_volume(context, volume, metadata, force):
            return metadata

        mock_copy_vol.side_effect = really_fake_upload_volume

        id = fake.VOLUME_ID
        img = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": img}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, id))
        res_dict = self.controller._volume_upload_image(req, id, body=body)

        sent_meta = res_dict['os-volume_upload_image']
        self.assertNotIn('cinder_encryption_key_id', sent_meta)
        self.assertNotIn('cinder_encryption_key_deletion_policy', sent_meta)

    def test_copy_volume_to_image_volumenotfound(self):
        def fake_volume_get_raise_exc(self, context, volume_id):
            raise exception.VolumeNotFound(volume_id=volume_id)

        self.mock_object(volume_api.API, 'get', fake_volume_get_raise_exc)

        id = fake.WILL_NOT_BE_FOUND_ID
        vol = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, id))
        self.assertRaises(exception.VolumeNotFound,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    @mock.patch.object(volume_api.API, 'copy_volume_to_image',
                       side_effect=exception.InvalidVolume(reason='blah'))
    def test_copy_volume_to_image_invalidvolume(self, mock_copy):
        id = fake.VOLUME2_ID
        vol = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get)
    def test_copy_volume_to_image_invalid_disk_format(self):
        id = fake.IMAGE_ID
        vol = {"container_format": 'bare',
               "disk_format": 'iso',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action'
                                      % (fake.PROJECT_ID, id))
        self.assertRaises(exception.ValidationError,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    def test_copy_volume_to_image_bad_disk_format_for_encrypted_vol(self):
        id = ENCRYPTED_VOLUME_ID
        vol = {"container_format": 'bare',
               "disk_format": 'qcow2',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action'
                                      % (fake.PROJECT_ID, id))
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    def test_copy_volume_to_image_bad_container_format_for_encrypted_vol(self):
        id = ENCRYPTED_VOLUME_ID
        vol = {"container_format": 'ovf',
               "disk_format": 'raw',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action'
                                      % (fake.PROJECT_ID, id))
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    @mock.patch.object(volume_api.API, "copy_volume_to_image")
    def test_copy_volume_to_image_disk_format_ploop(self,
                                                    mock_copy_to_image):
        volume = utils.create_volume(self.context, metadata={'test': 'test'})

        img = {"container_format": 'bare',
               "disk_format": 'ploop',
               "image_name": 'image_name'}
        body = {"os-volume_upload_image": img}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, volume.id))

        image_metadata = {'container_format': 'bare',
                          'disk_format': 'ploop',
                          'name': 'image_name'}
        self.controller._volume_upload_image(req, volume.id, body=body)

        mock_copy_to_image.assert_called_once_with(
            req.environ['cinder.context'], volume, image_metadata, False)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    @mock.patch.object(volume_api.API, 'copy_volume_to_image',
                       side_effect=ValueError)
    def test_copy_volume_to_image_valueerror(self, mock_copy):
        id = fake.VOLUME2_ID
        vol = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, fake.VOLUME_ID))
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    @mock.patch.object(volume_api.API, 'copy_volume_to_image',
                       side_effect=messaging.RemoteError)
    def test_copy_volume_to_image_remoteerror(self, mock_copy):
        id = fake.VOLUME2_ID
        vol = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": 'image_name',
               "force": True}
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, id))
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    @mock.patch.object(volume_api.API, 'copy_volume_to_image',
                       side_effect=messaging.RemoteError)
    @ddt.data(
        ({"image_name": 'image_name', "protected": None},
         exception.ValidationError),
        ({"image_name": 'image_name', "protected": ' '},
         exception.ValidationError),
        ({"image_name": 'image_name', "protected": 'test'},
         exception.ValidationError),
        ({"image_name": 'image_name', "visibility": 'test'},
         exception.ValidationError),
        ({"image_name": 'image_name', "visibility": ' '},
         exception.ValidationError),
        ({"image_name": 'image_name', "visibility": None},
         exception.ValidationError))
    @ddt.unpack
    def test_copy_volume_to_image_invalid_request_body(
            self, vol, exception, mock_copy):
        id = fake.VOLUME2_ID
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, id))
        req.api_version_request = api_version.APIVersionRequest("3.1")
        self.assertRaises(exception,
                          self.controller._volume_upload_image,
                          req, id, body=body)

    def test_volume_upload_image_typeerror(self):
        id = fake.VOLUME2_ID
        body = {"os-volume_upload_image_fake": "fake"}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, id))
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    def test_volume_upload_image_without_type(self):
        id = fake.VOLUME2_ID
        vol = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": None,
               "force": True}
        body = {"": vol}
        req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                  (fake.PROJECT_ID, id))
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(fake_auth_context=self.context))
        self.assertEqual(HTTPStatus.BAD_REQUEST, res.status_int)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get)
    def test_extend_volume_valueerror(self):
        id = fake.VOLUME2_ID
        body = {'os-extend': {'new_size': 'fake'}}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, id))
        self.assertRaises(exception.ValidationError,
                          self.controller._extend,
                          req,
                          id,
                          body=body)

    @ddt.data({'version': mv.get_prior_version(mv.VOLUME_EXTEND_INUSE),
               'status': 'available'},
              {'version': mv.get_prior_version(mv.VOLUME_EXTEND_INUSE),
               'status': 'in-use'},
              {'version': mv.VOLUME_EXTEND_INUSE,
               'status': 'available'},
              {'version': mv.VOLUME_EXTEND_INUSE,
               'status': 'in-use'})
    @ddt.unpack
    def test_extend_attached_volume(self, version, status):
        vol = db.volume_create(self.context,
                               {'size': 1, 'project_id': fake.PROJECT_ID,
                                'status': status,
                                'volume_type_id': fake.VOLUME_TYPE_ID})
        self.mock_object(volume_api.API, 'get', return_value=vol)
        mock_extend = self.mock_object(volume_api.API, '_extend')
        body = {"os-extend": {"new_size": 2}}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, vol['id']))
        req.api_version_request = mv.get_api_version(version)
        self.controller._extend(req, vol['id'], body=body)
        if version == mv.VOLUME_EXTEND_INUSE and status == 'in-use':
            mock_extend.assert_called_with(req.environ['cinder.context'],
                                           vol, 2, attached=True)
        else:
            mock_extend.assert_called_with(req.environ['cinder.context'],
                                           vol, 2, attached=False)

    def test_extend_volume_no_exist(self):
        vol_id = fake.WILL_NOT_BE_FOUND_ID

        body = {'os-extend': {'new_size': 5}}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, vol_id))

        self.assertRaises(exception.VolumeNotFound,
                          self.controller._extend,
                          req,
                          vol_id,
                          body=body)

    def test_copy_volume_to_image_notimagename(self):
        id = fake.VOLUME2_ID
        vol = {"container_format": 'bare',
               "disk_format": 'raw',
               "image_name": None,
               "force": True}
        body = {"os-volume_upload_image": vol}
        req = fakes.HTTPRequest.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, id))
        self.assertRaises(exception.ValidationError,
                          self.controller._volume_upload_image,
                          req,
                          id,
                          body=body)

    def _create_volume_with_type(self, status='available',
                                 display_description='displaydesc', **kwargs):
        admin_ctxt = context.get_admin_context()
        vol_type = db.volume_type_create(admin_ctxt, {'name': 'vol_name'})
        self.addCleanup(db.volume_type_destroy, admin_ctxt, vol_type.id)

        volume = utils.create_volume(self.context, volume_type_id=vol_type.id,
                                     status=status,
                                     display_description=display_description,
                                     **kwargs)
        self.addCleanup(db.volume_destroy, admin_ctxt, volume.id)

        expected = {
            'os-volume_upload_image': {
                'id': volume.id,
                'updated_at': mock.ANY,
                'status': 'uploading',
                'display_description': 'displaydesc',
                'size': 1,
                'volume_type': mock.ANY,
                'image_id': fake.IMAGE_ID,
                'container_format': 'bare',
                'disk_format': 'raw',
                'image_name': 'image_name'
            }
        }
        return volume, expected

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_with_protected_prop(
            self, mock_copy_to_image, mock_create, mock_get_image_metadata):
        """Test create image from volume with protected properties."""
        volume, expected = self._create_volume_with_type()
        mock_get_image_metadata.return_value = {"volume_id": volume.id,
                                                "key": "x_billing_license",
                                                "value": "246254365"}
        mock_create.side_effect = self.fake_image_service_create

        req = fakes.HTTPRequest.blank(
            '/v3/%s/volumes/%s/action' % (fake.PROJECT_ID, volume.id),
            use_admin_context=self.context.is_admin)
        body = self._get_os_volume_upload_image()

        res_dict = self.controller._volume_upload_image(req, volume.id,
                                                        body=body)

        self.assertDictEqual(expected, res_dict)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('uploading', vol_db.status)
        self.assertEqual('available', vol_db.previous_status)

    @mock.patch.object(volume_api.API, 'get', fake_volume_get_obj)
    def test_copy_volume_to_image_public_not_authorized(self):
        """Test unauthorized create public image from volume."""
        id = 'aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee'
        req = fakes.HTTPRequest.blank('/v3/tenant1/volumes/%s/action' % id)
        req.environ['cinder.context'].is_admin = False
        req.headers = mv.get_mv_header(mv.UPLOAD_IMAGE_PARAMS)
        req.api_version_request = mv.get_api_version(mv.UPLOAD_IMAGE_PARAMS)
        body = self._get_os_volume_upload_image()
        body['os-volume_upload_image']['visibility'] = 'public'
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller._volume_upload_image,
                          req, id, body=body)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_without_glance_metadata(
            self, mock_copy_to_image, mock_create, mock_get_image_metadata):
        """Test create image from volume if volume is created without image.

        In this case volume glance metadata will not be available for this
        volume.
        """
        volume, expected = self._create_volume_with_type()

        mock_get_image_metadata.side_effect = \
            exception.GlanceMetadataNotFound(id=volume.id)
        mock_create.side_effect = self.fake_image_service_create

        req = fakes.HTTPRequest.blank(
            '/v3/%s/volumes/%s/action' % (fake.PROJECT_ID, volume.id),
            use_admin_context=self.context.is_admin)
        body = self._get_os_volume_upload_image()
        res_dict = self.controller._volume_upload_image(req, volume.id,
                                                        body=body)

        self.assertDictEqual(expected, res_dict)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('uploading', vol_db.status)
        self.assertEqual('available', vol_db.previous_status)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_fail_image_create(
            self, mock_copy_to_image, mock_create, mock_get_image_metadata):
        """Test create image from volume if create image fails.

        In this case API will rollback to previous status.
        """
        volume = utils.create_volume(self.context)

        mock_get_image_metadata.return_value = {}
        mock_create.side_effect = Exception()

        req = fakes.HTTPRequest.blank(
            '/v3/fakeproject/volumes/%s/action' % volume.id)
        body = self._get_os_volume_upload_image()
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image, req, volume.id,
                          body=body)

        self.assertFalse(mock_copy_to_image.called)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('available', vol_db.status)
        self.assertIsNone(vol_db.previous_status)
        db.volume_destroy(context.get_admin_context(), volume.id)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_in_use_no_force(
            self, mock_copy_to_image, mock_create, mock_get_image_metadata):
        """Test create image from in-use volume.

        In this case API will fail because we are not passing force.
        """
        volume = utils.create_volume(self.context, status='in-use')

        mock_get_image_metadata.return_value = {}
        mock_create.side_effect = self.fake_image_service_create

        req = fakes.HTTPRequest.blank(
            '/v3/fakeproject/volumes/%s/action' % volume.id)
        body = self._get_os_volume_upload_image()
        body['os-volume_upload_image']['force'] = False
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image, req, volume.id,
                          body=body)

        self.assertFalse(mock_copy_to_image.called)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('in-use', vol_db.status)
        self.assertIsNone(vol_db.previous_status)
        db.volume_destroy(context.get_admin_context(), volume.id)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_in_use_with_force(
            self, mock_copy_to_image, mock_create, mock_get_image_metadata):
        """Test create image from in-use volume.

        In this case API will succeed only when CON.enable_force_upload is
        enabled.
        """
        volume, expected = self._create_volume_with_type(status='in-use')
        mock_get_image_metadata.return_value = {}
        mock_create.side_effect = self.fake_image_service_create

        req = fakes.HTTPRequest.blank(
            '/v3/fakeproject/volumes/%s/action' % volume.id,
            use_admin_context=self.context.is_admin)
        body = self._get_os_volume_upload_image()
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._volume_upload_image, req, volume.id,
                          body=body)

        self.assertFalse(mock_copy_to_image.called)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('in-use', vol_db.status)
        self.assertIsNone(vol_db.previous_status)

        CONF.set_default('enable_force_upload', True)
        res_dict = self.controller._volume_upload_image(req, volume.id,
                                                        body=body)

        self.assertDictEqual(expected, res_dict)

        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('uploading', vol_db.status)
        self.assertEqual('in-use', vol_db.previous_status)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_without_protected_prop(
            self, mock_volume_to_image, mock_create, mock_get_image_metadata):
        """Test protected property is not defined with the root image."""
        volume, expected = self._create_volume_with_type()

        mock_get_image_metadata.return_value = {}
        mock_create.side_effect = self.fake_image_service_create

        req = fakes.HTTPRequest.blank(
            '/v3/fakeproject/volumes/%s/action' % volume.id,
            use_admin_context=self.context.is_admin)

        body = self._get_os_volume_upload_image()
        res_dict = self.controller._volume_upload_image(req, volume.id,
                                                        body=body)

        self.assertDictEqual(expected, res_dict)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('uploading', vol_db.status)
        self.assertEqual('available', vol_db.previous_status)

    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_without_core_prop(
            self, mock_copy_to_image, mock_create):
        """Test glance_core_properties defined in cinder.conf is empty."""
        volume, expected = self._create_volume_with_type()
        mock_create.side_effect = self.fake_image_service_create

        self.override_config('glance_core_properties', [])

        req = fakes.HTTPRequest.blank(
            '/v3/fakeproject/volumes/%s/action' % volume.id,
            use_admin_context=self.context.is_admin)

        body = self._get_os_volume_upload_image()
        res_dict = self.controller._volume_upload_image(req, volume.id,
                                                        body=body)

        self.assertDictEqual(expected, res_dict)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('uploading', vol_db.status)
        self.assertEqual('available', vol_db.previous_status)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_volume_type_none(
            self,
            mock_copy_volume_to_image,
            mock_create,
            mock_get_volume_image_metadata):
        """Test create image from volume with none type volume."""
        volume, expected = self._create_volume_with_type()

        mock_create.side_effect = self.fake_image_service_create

        req = fakes.HTTPRequest.blank(
            '/v3/%s/volumes/%s/action' % (fake.PROJECT_ID, volume.id),
            use_admin_context=self.context.is_admin)
        body = self._get_os_volume_upload_image()
        res_dict = self.controller._volume_upload_image(req, volume.id,
                                                        body=body)
        self.assertDictEqual(expected, res_dict)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_version_with_params(
            self,
            mock_copy_volume_to_image,
            mock_create,
            mock_get_volume_image_metadata):
        """Test create image from volume with protected properties."""
        volume, expected = self._create_volume_with_type()

        mock_get_volume_image_metadata.return_value = {
            "volume_id": volume.id,
            "key": "x_billing_code_license",
            "value": "246254365"}
        mock_create.side_effect = self.fake_image_service_create_with_params
        mock_copy_volume_to_image.side_effect = \
            self.fake_rpc_copy_volume_to_image

        req = fakes.HTTPRequest.blank(
            '/v3/%s/volumes/%s/action' % (fake.PROJECT_ID, volume.id),
            use_admin_context=self.context.is_admin)
        req.environ['cinder.context'].is_admin = True
        req.headers = mv.get_mv_header(mv.UPLOAD_IMAGE_PARAMS)
        req.api_version_request = mv.get_api_version(mv.UPLOAD_IMAGE_PARAMS)
        body = self._get_os_volume_upload_image()
        body = self._get_os_volume_upload_image()
        body['os-volume_upload_image']['visibility'] = 'public'
        body['os-volume_upload_image']['protected'] = True
        res_dict = self.controller._volume_upload_image(req,
                                                        volume.id,
                                                        body=body)

        expected['os-volume_upload_image'].update(visibility='public',
                                                  protected=True)
        self.assertDictEqual(expected, res_dict)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_vhd(
            self, mock_copy_to_image, mock_create, mock_get_image_metadata):
        """Test create image from volume with vhd disk format"""
        volume, expected = self._create_volume_with_type()
        mock_get_image_metadata.return_value = {}
        mock_create.side_effect = self.fake_image_service_create
        req = fakes.HTTPRequest.blank(
            '/v3/fakeproject/volumes/%s/action' % volume.id)
        body = self._get_os_volume_upload_image()
        body['os-volume_upload_image']['force'] = True
        body['os-volume_upload_image']['container_format'] = 'bare'
        body['os-volume_upload_image']['disk_format'] = 'vhd'

        res_dict = self.controller._volume_upload_image(req, volume.id,
                                                        body=body)

        self.assertDictEqual(expected, res_dict)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('uploading', vol_db.status)
        self.assertEqual('available', vol_db.previous_status)

    @mock.patch.object(volume_api.API, "get_volume_image_metadata")
    @mock.patch.object(glance.GlanceImageService, "create")
    @mock.patch.object(volume_rpcapi.VolumeAPI, "copy_volume_to_image")
    def test_copy_volume_to_image_vhdx(
            self, mock_copy_to_image, mock_create, mock_get_image_metadata):
        """Test create image from volume with vhdx disk format"""
        volume, expected = self._create_volume_with_type()
        mock_get_image_metadata.return_value = {}
        mock_create.side_effect = self.fake_image_service_create
        req = fakes.HTTPRequest.blank(
            '/v3/fakeproject/volumes/%s/action' % volume.id)
        body = self._get_os_volume_upload_image()
        body['os-volume_upload_image']['force'] = True
        body['os-volume_upload_image']['container_format'] = 'bare'
        body['os-volume_upload_image']['disk_format'] = 'vhdx'

        res_dict = self.controller._volume_upload_image(req, volume.id,
                                                        body=body)

        self.assertDictEqual(expected, res_dict)
        vol_db = objects.Volume.get_by_id(self.context, volume.id)
        self.assertEqual('uploading', vol_db.status)
        self.assertEqual('available', vol_db.previous_status)

    def _build_reimage_req(self, body, vol_id,
                           version=mv.SUPPORT_REIMAGE_VOLUME):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/volumes/%s/action' % (fake.PROJECT_ID, id))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = self.context
        req.api_version_request = mv.get_api_version(version)
        req.headers["content-type"] = "application/json"
        return req

    @ddt.data(None, False, True)
    @mock.patch.object(volume_api.API, "reimage")
    def test_volume_reimage(self, reimage_reserved, mock_image):
        vol = utils.create_volume(self.context)
        body = {"os-reimage": {"image_id": fake.IMAGE_ID}}
        if reimage_reserved is not None:
            body["os-reimage"]["reimage_reserved"] = reimage_reserved
        req = self._build_reimage_req(body, vol.id)
        self.controller._reimage(req, vol.id, body=body)

    @mock.patch.object(volume_api.API, "reimage")
    def test_volume_reimage_invaild_params(self, mock_image):
        vol = utils.create_volume(self.context)
        body = {"os-reimage": {"image_id": fake.IMAGE_ID,
                               "reimage_reserved": 'wrong'}}
        req = self._build_reimage_req(body, vol)
        self.assertRaises(exception.ValidationError,
                          self.controller._reimage, req,
                          vol.id, body=body)

    def test_volume_reimage_before_3_68(self):
        vol = utils.create_volume(self.context)
        body = {"os-reimage": {"image_id": fake.IMAGE_ID}}

        req = self._build_reimage_req(body, vol.id, version="3.67")
        self.assertRaises(exception.VersionNotFoundForAPIMethod,
                          self.controller._reimage, req, vol.id, body=body)

    def test_reimage_volume_invalid_status(self):
        def fake_reimage_volume(*args, **kwargs):
            msg = "Volume status must be available."
            raise exception.InvalidVolume(reason=msg)
        self.mock_object(volume.api.API, 'reimage',
                         fake_reimage_volume)

        vol = utils.create_volume(self.context)
        body = {"os-reimage": {"image_id": fake.IMAGE_ID}}
        req = self._build_reimage_req(body, vol)
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._reimage, req,
                          vol.id, body=body)

    @mock.patch('cinder.context.RequestContext.authorize')
    def test_reimage_volume_attach_more_than_one_server(self, mock_authorize):
        vol = utils.create_volume(self.context)
        va_objs = [objects.VolumeAttachment(context=self.context, id=i)
                   for i in [fake.OBJECT_ID, fake.OBJECT2_ID, fake.OBJECT3_ID]]
        va_list = objects.VolumeAttachmentList(context=self.context,
                                               objects=va_objs)
        vol.volume_attachment = va_list
        self.mock_object(volume_api.API, 'get', return_value=vol)
        body = {"os-reimage": {"image_id": fake.IMAGE_ID}}
        req = self._build_reimage_req(body, vol)
        self.assertRaises(webob.exc.HTTPConflict,
                          self.controller._reimage, req, vol.id, body=body)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from cinder.api.contrib import quotas
from cinder.api import microversions as mv
from cinder.policies import quotas as policy
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit.policies import base


@ddt.ddt
class QuotasPolicyTest(base.BasePolicyTest):
    authorized_users = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_users = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_admins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]

    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    unauthorized_exceptions = []

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = quotas.QuotaSetsController()
        self.api_path = '/v3/os-quota-sets'
        self.api_version = mv.BASE_VERSION

    @ddt.data(*base.all_users)
    def test_show_policy(self, user_id):
        rule_name = policy.SHOW_POLICY
        req = fake_api.HTTPRequest.blank(self.api_path,
                                         version=self.api_version)

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.show,
                                 req, id=self.project_id)

    @ddt.data(*base.all_users)
    def test_update_policy(self, user_id):
        rule_name = policy.UPDATE_POLICY
        req = fake_api.HTTPRequest.blank(self.api_path,
                                         version=self.api_version)
        req.method = 'PUT'
        body = {
            "quota_set": {
                "groups": 11,
                "volumes": 5,
                "backups": 4
            }
        }

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.update,
                                 req, id=self.project_id, body=body)

    @ddt.data(*base.all_users)
    def test_delete_policy(self, user_id):
        rule_name = policy.DELETE_POLICY
        req = fake_api.HTTPRequest.blank(self.api_path,
                                         version=self.api_version)
        req.method = 'DELETE'

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.delete,
                                 req, id=self.project_id)


class QuotasPolicySecureRbacTest(QuotasPolicyTest):
    authorized_users = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
    ]

    unauthorized_users = [
        'legacy_owner',
        'system_member',
        'system_foo',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # NOTE(Xena): The authorized_admins and unauthorized_admins are the same
    # as the QuotasPolicyTest's. This is because in Xena the "admin only"
    # rules are the legacy RULE_ADMIN_API. This will change in Yoga, when
    # RULE_ADMIN_API will be deprecated in favor of the SYSTEM_ADMIN rule that
    # is scope based.

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



# Copyright 2016 Intel Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for volume type."""

from cinder import context
from cinder import db
from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import utils
from cinder.volume import volume_types


class VolumeTypeTestCase(test.TestCase):
    """Test cases for volume type."""

    def setUp(self):
        super(VolumeTypeTestCase, self).setUp()
        self.ctxt = context.RequestContext(user_id=fake.USER_ID,
                                           project_id=fake.PROJECT_ID,
                                           is_admin=True)

    def test_volume_type_delete(self):
        volume_type = db.volume_type_create(self.ctxt, {'name':
                                                        'fake volume type'})
        volume_types.destroy(self.ctxt, volume_type['id'])
        self.assertRaises(exception.VolumeTypeNotFound,
                          volume_types.get_by_name_or_id, self.ctxt,
                          volume_type['id'])

    def test_volume_db_delete_last_type(self):
        default = volume_types.get_default_volume_type()
        self.assertRaises(exception.VolumeTypeDeletionError,
                          db.volume_type_destroy, self.ctxt,
                          default['id'])

    def test_volume_type_delete_with_volume_in_use(self):
        volume_type = db.volume_type_create(self.ctxt, {'name':
                                                        'fake volume type'})
        volume = db.volume_create(self.ctxt, {'volume_type_id':
                                              volume_type['id']})
        self.assertRaises(exception.VolumeTypeInUse, volume_types.destroy,
                          self.ctxt, volume_type['id'])
        db.volume_destroy(self.ctxt, volume['id'])
        volume_types.destroy(self.ctxt, volume_type['id'])

    def test_volume_type_delete_with_group_in_use(self):
        volume_type = db.volume_type_create(self.ctxt, {'name':
                                                        'fake volume type'})

        group = db.group_create(self.ctxt, {})
        db.group_volume_type_mapping_create(self.ctxt, group['id'],
                                            volume_type['id'])
        self.assertRaises(exception.VolumeTypeInUse, volume_types.destroy,
                          self.ctxt, volume_type['id'])
        db.group_destroy(self.ctxt, group['id'])
        volume_types.destroy(self.ctxt, volume_type['id'])

    def test_volume_type_mark_in_use_exists(self):
        volume_type = db.volume_type_create(
            self.ctxt,
            {'name': 'fake volume type'},
        )
        group = db.group_create(self.ctxt, {})
        db.group_volume_type_mapping_create(
            self.ctxt,
            group['id'],
            volume_type['id'],
        )
        self.assertRaises(
            exception.GroupVolumeTypeMappingExists,
            db.group_volume_type_mapping_create,
            self.ctxt,
            group['id'],
            volume_type['id'],
        )

    def test_volume_type_delete_with_consistencygroups_in_use(self):
        volume_type = db.volume_type_create(self.ctxt, {'name':
                                                        'fake volume type'})
        consistency_group1 = db.consistencygroup_create(self.ctxt,
                                                        {'volume_type_id':
                                                         volume_type['id']})
        consistency_group2 = db.consistencygroup_create(self.ctxt,
                                                        {'volume_type_id':
                                                         volume_type['id']})
        self.assertRaises(exception.VolumeTypeInUse, volume_types.destroy,
                          self.ctxt, volume_type['id'])
        db.consistencygroup_destroy(self.ctxt, consistency_group1['id'])
        self.assertRaises(exception.VolumeTypeInUse, volume_types.destroy,
                          self.ctxt, volume_type['id'])
        db.consistencygroup_destroy(self.ctxt, consistency_group2['id'])
        volume_types.destroy(self.ctxt, volume_type['id'])

    def test_volume_type_update(self):
        vol_type_ref = volume_types.create(self.ctxt, 'fake volume type')
        updates = dict(name='test_volume_type_update',
                       description=None,
                       is_public=None)
        db.volume_type_update(self.ctxt, vol_type_ref.id, updates)
        updated_vol_type = db.volume_type_get(self.ctxt, vol_type_ref.id)
        self.assertEqual('test_volume_type_update', updated_vol_type['name'])
        volume_types.destroy(self.ctxt, vol_type_ref.id)

    def test_volume_type_get_with_qos_specs(self):
        """Ensure volume types get can load qos_specs."""
        qos_data = {'name': 'qos', 'consumer': 'front-end',
                    'specs': {'key': 'value', 'key2': 'value2'}}
        qos = utils.create_qos(self.ctxt, **qos_data)
        vol_type = db.volume_type_create(self.ctxt,
                                         {'name': 'my-vol-type',
                                          'qos_specs_id': qos['id']})

        db_vol_type = db.volume_type_get(self.ctxt, vol_type.id,
                                         expected_fields=['qos_specs'])

        expected = {('QoS_Specs_Name', 'qos'), ('consumer', 'front-end'),
                    ('key', 'value'), ('key2', 'value2')}
        actual = {(spec.key, spec.value) for spec in db_vol_type['qos_specs']}
        self.assertEqual(expected, actual)

    def test_volume_type_get_with_projects(self):
        """Ensure volume types get can load projects."""
        projects = [fake.PROJECT_ID, fake.PROJECT2_ID, fake.PROJECT3_ID]
        vol_type = db.volume_type_create(self.ctxt,
                                         {'name': 'my-vol-type'},
                                         projects=projects)

        db_vol_type = db.volume_type_get(self.ctxt, vol_type.id,
                                         expected_fields=['projects'])

        self.assertEqual(set(projects), set(db_vol_type['projects']))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for resource filters API."""

from unittest import mock

import ddt

from cinder.api import microversions as mv
from cinder.api.v3 import resource_filters as v3_filters
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test


@ddt.ddt
class ResourceFiltersAPITestCase(test.TestCase):
    """Test Case for filter API."""

    def setUp(self):
        super(ResourceFiltersAPITestCase, self).setUp()
        self.controller = v3_filters.ResourceFiltersController()

    @ddt.data({'filters': {'volume': ['key1']},
               'resource': 'volume',
               'expected_filters': [{'resource': 'volume',
                                     'filters': ['key1']}]},
              {'filters': {'volume': ['key1'], 'snapshot': ['key2']},
               'resource': None,
               'expected_filters': [{'resource': 'volume',
                                     'filters': ['key1']},
                                    {'resource': 'snapshot',
                                     'filters': ['key2']}]},
              {'filters': {'volume': ['key1', 'key2']},
               'resource': 'snapshot',
               'expected_filters': []})
    @ddt.unpack
    def test_get_allowed_filters(self, filters, resource, expected_filters):
        request_url = '/v3/%s/resource_filters' % fake.PROJECT_ID
        if resource is not None:
            request_url += '?resource=%s' % resource
        req = fakes.HTTPRequest.blank(request_url,
                                      version=mv.RESOURCE_FILTER_CONFIG)

        with mock.patch('cinder.api.common._FILTERS_COLLECTION', filters):
            result = self.controller.index(req)
            self.assertCountEqual(list(result), ['resource_filters'])
            self.assertCountEqual(expected_filters, result['resource_filters'])



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base
from cinder.volume import volume_types
from cinder.volume import volume_utils


@base.CinderObjectRegistry.register
class VolumeType(base.CinderPersistentObject, base.CinderObject,
                 base.CinderObjectDictCompat, base.CinderComparableObject):
    # Version 1.0: Initial version
    # Version 1.1: Changed extra_specs to DictOfNullableStringsField
    # Version 1.2: Added qos_specs
    # Version 1.3: Add qos_specs_id
    VERSION = '1.3'

    OPTIONAL_FIELDS = ('extra_specs', 'projects', 'qos_specs')

    # NOTE: When adding a field obj_make_compatible needs to be updated
    fields = {
        'id': fields.UUIDField(),
        'name': fields.StringField(nullable=True),
        'description': fields.StringField(nullable=True),
        'is_public': fields.BooleanField(default=True, nullable=True),
        'projects': fields.ListOfStringsField(nullable=True),
        'extra_specs': fields.DictOfNullableStringsField(nullable=True),
        'qos_specs_id': fields.UUIDField(nullable=True),
        'qos_specs': fields.ObjectField('QualityOfServiceSpecs',
                                        nullable=True),
    }

    @classmethod
    def _get_expected_attrs(cls, context, *args, **kwargs):
        return 'extra_specs', 'projects'

    @classmethod
    def _from_db_object(cls, context, type, db_type, expected_attrs=None):
        if expected_attrs is None:
            expected_attrs = ['extra_specs', 'projects']
        for name, field in type.fields.items():
            if name in cls.OPTIONAL_FIELDS:
                continue
            value = db_type[name]
            if isinstance(field, fields.IntegerField):
                value = value or 0
            type[name] = value

        # Get data from db_type object that was queried by joined query
        # from DB
        if 'extra_specs' in expected_attrs:
            type.extra_specs = {}
            specs = db_type.get('extra_specs')
            if specs and isinstance(specs, list):
                type.extra_specs = {item['key']: item['value']
                                    for item in specs}
            elif specs and isinstance(specs, dict):
                type.extra_specs = specs
        if 'projects' in expected_attrs:
            # NOTE(geguileo): Until projects stops being a polymorphic value we
            # have to do a conversion here for VolumeTypeProjects ORM instance
            # lists.
            projects = db_type.get('projects', [])
            if projects and not isinstance(projects[0], str):
                projects = [p.project_id for p in projects]
            type.projects = projects
        if 'qos_specs' in expected_attrs:
            qos_specs = objects.QualityOfServiceSpecs(context)
            qos_specs._from_db_object(context, qos_specs, db_type['qos_specs'])
            type.qos_specs = qos_specs
        type._context = context
        type.obj_reset_changes()
        return type

    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        db_volume_type = volume_types.create(self._context, self.name,
                                             self.extra_specs,
                                             self.is_public, self.projects,
                                             self.description)
        self._from_db_object(self._context, self, db_volume_type)

    def save(self):
        updates = self.cinder_obj_get_changes()
        if updates:
            volume_types.update(self._context, self.id, self.name,
                                self.description)
            self.obj_reset_changes()

    def destroy(self):
        with self.obj_as_admin():
            updated_values = volume_types.destroy(self._context, self.id)
        self.update(updated_values)
        self.obj_reset_changes(updated_values.keys())

    def obj_load_attr(self, attrname):
        if attrname not in self.OPTIONAL_FIELDS:
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        if attrname == 'extra_specs':
            self.extra_specs = db.volume_type_extra_specs_get(self._context,
                                                              self.id)

        elif attrname == 'qos_specs':
            if self.qos_specs_id:
                self.qos_specs = objects.QualityOfServiceSpecs.get_by_id(
                    self._context, self.qos_specs_id)
            else:
                self.qos_specs = None

        elif attrname == 'projects':
            volume_type_projects = db.volume_type_access_get_all(self._context,
                                                                 self.id)
            self.projects = [x.project_id for x in volume_type_projects]

        self.obj_reset_changes(fields=[attrname])

    @classmethod
    def get_by_name_or_id(cls, context, identity):
        orm_obj = volume_types.get_by_name_or_id(context, identity)
        expected_attrs = cls._get_expected_attrs(context)
        return cls._from_db_object(context, cls(context),
                                   orm_obj, expected_attrs=expected_attrs)

    def is_replicated(self):
        return volume_utils.is_replicated_spec(self.extra_specs)

    def is_multiattach(self):
        return volume_utils.is_multiattach_spec(self.extra_specs)


@base.CinderObjectRegistry.register
class VolumeTypeList(base.ObjectListBase, base.CinderObject):
    # Version 1.0: Initial version
    # Version 1.1: Add pagination support to volume type
    VERSION = '1.1'

    fields = {
        'objects': fields.ListOfObjectsField('VolumeType'),
    }

    @classmethod
    def get_all(cls, context, inactive=0, filters=None, marker=None,
                limit=None, sort_keys=None, sort_dirs=None, offset=None):
        types = volume_types.get_all_types(context, inactive, filters,
                                           marker=marker, limit=limit,
                                           sort_keys=sort_keys,
                                           sort_dirs=sort_dirs, offset=offset)
        expected_attrs = VolumeType._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context),
                                  objects.VolumeType, types.values(),
                                  expected_attrs=expected_attrs)

    @classmethod
    def get_all_types_for_qos(cls, context, qos_id):
        types = db.qos_specs_associations_get(context, qos_id)
        return base.obj_make_list(context, cls(context), objects.VolumeType,
                                  types)

    @classmethod
    def get_all_by_group(cls, context, group_id):
        # Generic volume group
        types = db.volume_type_get_all_by_group(
            context.elevated(), group_id)
        expected_attrs = VolumeType._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context),
                                  objects.VolumeType, types,
                                  expected_attrs=expected_attrs)



#  Copyright (C) 2021-2022 YADRO.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import time

from oslo_log import log as logging
import requests

from cinder import exception
from cinder.i18n import _
from cinder.utils import retry
from cinder.volume.drivers.yadro import tatlin_api
from cinder.volume.drivers.yadro.tatlin_exception import TatlinAPIException

LOG = logging.getLogger(__name__)

retry_exc = (Exception,)


def InitTatlinClient(ip, port, username, password,
                     verify, api_retry_count,
                     wait_interval, wait_retry_count):
    access_api = TatlinAccessAPI(ip, port, username, password, verify)
    tatlin_version = access_api.get_tatlin_version()
    if tatlin_version <= (2, 3):
        return TatlinClientV23(access_api,
                               api_retry_count=api_retry_count,
                               wait_interval=wait_interval,
                               wait_retry_count=wait_retry_count)
    else:
        return TatlinClientV25(access_api,
                               api_retry_count=api_retry_count,
                               wait_interval=wait_interval,
                               wait_retry_count=wait_retry_count)


class TatlinAccessAPI:
    session = None
    ip = None
    port = None
    username = None
    password = None
    verify = False
    _api_version = None

    def __init__(self, ip, port, user, passwd, verify):
        self.ip = ip
        self.port = port
        self.username = user
        self.password = passwd
        self.verify = verify
        self._authenticate_access()

    def _authenticate_access(self):
        LOG.debug('Generating new Tatlin API session')

        self.session = requests.session()
        LOG.debug('SSL verification %s', self.session.verify)
        self.session.verify = self.verify
        if not self.verify:
            requests.packages.urllib3.disable_warnings()

        # Here 'address' will be only IPv4.
        response = self.session.post('https://%s:%d/auth/login'
                                     % (self.ip, self.port),
                                     data={'user': self.username,
                                           'secret': self.password},
                                     verify=self.verify)
        if response.status_code != requests.codes.ok:
            LOG.error('Failed to authenticate to remote cluster at %s for %s.',
                      self.ip, self.username)
            raise exception.NotAuthorized(_('Authentication failure.'))
        result = response.json()
        self.session.headers.update({'X-Auth-Token': result['token']})
        self.session.headers.update({'Content-Type': 'application/json'})

    def send_request(self, path, input_data, method):
        full_url = self._get_api(path)
        resp = self.session.request(
            method, full_url, verify=self.verify, json=input_data)
        LOG.debug('Tatlin response for method %s URL %s %s',
                  method, full_url, resp)
        if resp.status_code == requests.codes.unauthorized:
            LOG.info('Not authenticated. Logging in.')
            self._authenticate_access()
            resp = self.session.request(
                method, full_url, verify=self.verify, json=input_data)
        return resp

    def get_tatlin_version(self):
        if not self._api_version:
            responce = self.send_request(tatlin_api.TATLIN_VERSION,
                                         {}, 'GET')
            ver = responce.json()['build-version'].split('.')
            self._api_version = (int(ver[0]), int(ver[1]))
        LOG.debug('Tatlin version: %s', str(self._api_version))
        return self._api_version

    def _get_api(self, tail):
        return ('https://%s:%d/' % (self.ip, self.port)) + tail


class TatlinClientCommon:
    session = None
    _api = None
    access_api_retry_count = 1

    def __init__(self, tatlin_rest_api, api_retry_count,
                 wait_interval, wait_retry_count):
        self.session = None
        self._api = tatlin_rest_api
        self.access_api_retry_count = api_retry_count
        self.wait_interval = wait_interval
        self.wait_retry_count = wait_retry_count

    def add_vol_to_host(self, vol_id, host_id):
        LOG.debug('Adding volume %s to host %s', vol_id, host_id)
        if self._is_vol_on_host(vol_id, host_id):
            return
        path = tatlin_api.VOLUME_TO_HOST % (vol_id, host_id)
        try:
            self._access_api(path, {}, 'PUT',
                             pass_codes=[requests.codes.bad_request])
        except TatlinAPIException as exp:
            message = _('Unable to add volume %s to host %s error %s' %
                        (vol_id, host_id, exp.message))
            LOG.error(message)
            raise TatlinAPIException(500, message)

        if not self._is_vol_on_host(vol_id, host_id):
            raise exception.VolumeBackendAPIException(
                'Unable to add volume %s to host  %s' % (vol_id, host_id))
        return

    def remove_vol_from_host(self, vol_id, host_id):
        if not self._is_vol_on_host(vol_id, host_id):
            return
        path = tatlin_api.VOLUME_TO_HOST % (vol_id, host_id)
        try:
            LOG.debug('Removing volume %s from host %s', vol_id, host_id)
            self._access_api(path, {}, 'DELETE',
                             pass_codes=[requests.codes.not_found,
                                         requests.codes.bad_request])
        except TatlinAPIException as exp:
            message = _('Unable to remove volume %s from host %s error %s' %
                        (vol_id, host_id, exp.message))
            LOG.error(message)
            raise TatlinAPIException(500, message)

        if self._is_vol_on_host(vol_id, host_id):
            raise exception.VolumeBackendAPIException(
                'Volume %s still on host  %s' % (vol_id, host_id))
        return

    def create_volume(self,
                      vol_id, name,
                      size_in_byte,
                      pool_id,
                      lbaFormat='512e'):

        data = {"name": name,
                "size": size_in_byte,
                "poolId": pool_id,
                "deduplication": False,
                "compression": False,
                "alert_threshold": 0,
                "lbaFormat": lbaFormat
                }
        path = tatlin_api.RESOURCE % vol_id
        LOG.debug('Create volume: volume=%(v3)s path=%(v1)s body=%(v2)s',
                  {'v1': path, 'v2': data, 'v3': vol_id},)

        try:
            self._access_api(path, data, 'PUT')
        except TatlinAPIException as exp:
            message = _('Create volume %s failed due to %s' %
                        (id, exp.message))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def delete_volume(self, vol_id):
        LOG.debug('Delete volume %s', vol_id)
        path = tatlin_api.RESOURCE % vol_id
        try:
            self._access_api(path, {}, 'DELETE',
                             pass_codes=[requests.codes.not_found,
                                         requests.codes.bad_request])
        except TatlinAPIException as exp:
            message = _('Delete volume %s failed due to %s' %
                        (vol_id, exp.message))
            LOG.error(message)
            raise

    def extend_volume(self, vol_id, new_size_in_byte):
        path = tatlin_api.RESOURCE % vol_id
        data = {"new_size": new_size_in_byte}
        LOG.debug('Extending volume to %s ', new_size_in_byte)
        try:
            self._access_api(path, data, 'POST')
        except TatlinAPIException as exp:
            message = _('Unable to extend volume %s due to %s' %
                        (vol_id, exp.message))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def get_resource_mapping(self):
        try:
            result, status = self._access_api(tatlin_api.RESOURCE_MAPPING)
            return result
        except TatlinAPIException as exp:
            message = _(
                'TATLIN: Error getting resource mapping information %s' %
                exp.message)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def get_all_hosts(self):
        try:
            result, status = self._access_api(tatlin_api.HOSTS)
            return result
        except TatlinAPIException:
            message = _('Unable to get hosts configuration')
            raise exception.VolumeBackendAPIException(message=message)

    def get_host_info(self, host_id):
        try:
            result, stat = self._access_api(tatlin_api.HOSTS + '/' + host_id)
            LOG.debug('Host info for %s is %s', host_id, result)
            return result
        except TatlinAPIException as exp:
            message = _('Unable to get host info %s error %s' %
                        (host_id, exp.message))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def get_host_id(self, name):
        return self.get_host_id_by_name(name)

    def get_iscsi_cred(self):
        auth_path = tatlin_api.RESOURCE % 'auth'
        try:
            cred, status = self._access_api(auth_path)
        except TatlinAPIException as exp:
            message = _('Unable to get iscsi user cred due to %s' %
                        exp.message)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)
        return cred

    def get_host_group_info(self, group_id):
        try:
            result, status = self._access_api(tatlin_api.HOST_GROUPS + '/' +
                                              group_id)
            return result
        except TatlinAPIException as exp:
            message = _('Unable to get host group info %s error %s' %
                        (group_id, exp.message))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def get_host_group_id(self, name):
        try:
            result, status = self._access_api(tatlin_api.HOST_GROUPS)
            for h in result:
                LOG.debug('Host name: %s Host ID %s', h['name'], h['id'])
                if h['name'] == name:
                    return h['id']
        except TatlinAPIException as exp:
            message = (_('Unable to get id for host group %s error %s') %
                       (name, exp.message))
            LOG.error(message)
        raise exception.VolumeBackendAPIException(
            message='Unable to find host group id for %s' % name)

    def get_volume_ports(self, vol_id):
        if not self.is_volume_exists(vol_id):
            message = _('Unable to get volume info %s' % vol_id)
            LOG.error(message)
            return {}
        path = tatlin_api.RESOURCE % vol_id + '/ports'
        try:
            response, stat = self._access_api(path)
        except TatlinAPIException as exp:
            message = _('Unable to get ports for target %s '
                        'with %s error code: %s' %
                        (vol_id, exp.message, exp.code))
            LOG.error(message)
            return {}
        return response

    def get_resource_ports_array(self, volume_id):
        ports = self.get_volume_ports(volume_id)
        if ports == {}:
            return []
        res = []
        for p in ports:
            res.append(p['port'])
        LOG.debug('Volume %s port list %s', volume_id, res)
        return res

    def get_port_portal(self, portal_type):
        path = tatlin_api.IP_PORTS % portal_type
        try:
            result, stat = self._access_api(path)
        except TatlinAPIException as exp:
            message = _('Failed to get ports info due to %s' % exp.message)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)
        return result

    def is_volume_exists(self, vol_id):
        volume_path = tatlin_api.RESOURCE % vol_id
        LOG.debug('get personality statistic: volume_path=%(v1)s ',
                  {'v1': volume_path})
        try:
            volume_result, status = self._access_api(
                volume_path, {}, 'GET',
                pass_codes=[requests.codes.not_found])
            if status == requests.codes.not_found:
                message = _('Volume %s does not exist' % vol_id)
                LOG.debug(message)
                return False
        except TatlinAPIException as exp:
            message = _('Exception Unable to get volume info %s '
                        'due to %s stat: %s' %
                        (vol_id, exp.message, exp.code))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)
        LOG.debug('Volume %s exists', vol_id)
        return True

    def get_volume(self, vol_id):
        volume_path = tatlin_api.RESOURCE % vol_id
        LOG.debug('get personality statistic: volume_path=%(v1)s',
                  {'v1': volume_path})
        try:
            volume_result, stat = self._access_api(
                volume_path, {}, 'GET',
                pass_codes=[requests.codes.not_found])
            if stat == requests.codes.not_found:
                message = _('Unable to get volume info %s due to %s stat: %s' %
                            (vol_id, 'Volume not found', '404'))
                LOG.error(message)
                raise exception.VolumeBackendAPIException(message=message)
        except TatlinAPIException as exp:
            message = _('Unable to get volume info %s due to %s stat: %s' %
                        (vol_id, exp.message, exp.code))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)
        return volume_result

    def get_pool_id_by_name(self, pool_name):
        try:
            result, status = self._access_api(tatlin_api.POOLS)
        except TatlinAPIException as exp:
            message = _('Unable to get pool id for %s due to %s' %
                        pool_name, exp.message)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        for p in result:
            if p['name'] == pool_name:
                return p['id']

        message = _('Pool "%s" not found' % pool_name)
        LOG.error(message)
        raise exception.VolumeBackendAPIException(message=message)

    def get_pool_detail(self, pool_id):
        if not pool_id:
            return {}
        path = tatlin_api.POOLS + "/" + pool_id
        try:
            result, status = self._access_api(path)
        except TatlinAPIException as exp:
            message = _('Unable to get pool information for %s due to %s' %
                        (pool_id, exp.message))
            LOG.error(message)
            return {}
        return result

    def get_sys_statistic(self):
        try:
            sys_stat, status = self._access_api(tatlin_api.STATISTICS)
        except TatlinAPIException as exp:
            message = _('Unable to get system statistic due to %s' %
                        exp.message)
            LOG.error(message)
            raise
        return sys_stat

    def get_volume_info(self, vol_name):
        path = tatlin_api.RESOURCE_DETAIL % vol_name
        try:
            result, status = self._access_api(path)
        except TatlinAPIException as exp:
            message = _('Unable to get volume %s error %s' %
                        (vol_name, exp.message))
            LOG.error(message)
            raise exception.ManageExistingInvalidReference(message)

        return result

    def get_tatlin_version(self):
        return self._api.get_tatlin_version()

    def get_resource_count(self, p_id):
        raise NotImplementedError()

    def is_volume_ready(self, id):
        path = tatlin_api.RESOURCE_DETAIL % id
        try:
            result, status = self._access_api(path)
        except TatlinAPIException:
            return False

        for p in result:
            LOG.debug('Volume %s status: %s', id, p['status'])
            if p['status'] != 'ready':
                return False

        return True

    def get_volume_status(self, id):
        path = tatlin_api.RESOURCE_HEALTH % id
        try:
            result, status = self._access_api(path)
        except TatlinAPIException:
            return False

        for p in result:
            LOG.debug('Volume status: %s', p['status'])
            return p['status']

        return ''

    def set_port(self, vol_id, port):
        path = tatlin_api.RESOURCE % vol_id + "/ports/" + port
        try:
            self._access_api(path, {}, 'PUT',
                             pass_codes=[requests.codes.conflict])
        except TatlinAPIException as e:
            message = _('Unable to link port %s for volume %s error %s' %
                        (port, vol_id, e.message))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def export_volume(self, vol_id, eth_ports):
        raise NotImplementedError()

    def export_vol_to_port_list(self, vol_id, port_list):
        path = tatlin_api.RESOURCE % vol_id + "/ports/list"
        try:
            self._access_api(path,
                             port_list, 'PUT',
                             pass_codes=[
                                 requests.codes.conflict,
                                 requests.codes.bad_request])
        except TatlinAPIException as e:
            message = _('Unable to link ports %s for volume %s error %s' %
                        (port_list, vol_id, e.message))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def _access_api(self, path, input_data = None, method = None,
                    pass_codes=None):
        @retry(retry_exc, interval=1,
               retries=self.access_api_retry_count)
        def do_access_api(path, input_data, method,
                          pass_codes):
            if input_data is None:
                input_data = {}
            if method is None:
                method = 'GET'
            if pass_codes is None:
                pass_codes = []
            pass_codes = [requests.codes.ok] + pass_codes
            startTime = time.time()
            response = self._api.send_request(path, input_data, method)
            finishTime = time.time()
            duration = str((finishTime - startTime) * 1000) + ' ms'
            postfix = '[FAST]' if finishTime - startTime < 15 else '[SLOW]'
            try:
                result = response.json()
            except ValueError:
                result = {}
            if response.status_code not in pass_codes:
                message = _('Request: method: %s path: %s '
                            'failed with status: %s message: %s in %s %s' %
                            (method, path, str(response.status_code),
                             result, duration, postfix))

                LOG.debug(message)
                raise TatlinAPIException(response.status_code,
                                         message, path=path)
            LOG.debug(
                'Request %s %s successfully finished with %s code in %s %s',
                method, path, str(response.status_code), duration, postfix)
            return result, response.status_code
        return do_access_api(path, input_data, method,
                             pass_codes)

    def _is_vol_on_host(self, vol_id, host_id):
        LOG.debug('Check resource %s in host %s', vol_id, host_id)
        try:
            result, status = self._access_api(tatlin_api.RESOURCE_MAPPING)
        except TatlinAPIException as exp:
            raise exception.VolumeBackendAPIException(
                message=_('Tatlin API exception %s '
                          'while getting resource mapping' % exp.message))

        for entry in result:
            if 'host_id' in entry:
                if entry['resource_id'] == vol_id and \
                        entry['host_id'] == host_id:
                    LOG.debug('Volume %s already on host %s',
                              vol_id, host_id)
                    return True
        LOG.debug('Volume %s not on host %s', vol_id, host_id)
        return False

    def get_unassigned_ports(self, volume_id, eth_ports):
        cur_ports = self.get_resource_ports_array(volume_id)
        LOG.debug('VOLUME %s: Port needed %s actual %s',
                  volume_id, list(eth_ports.keys()), cur_ports)
        return list(set(eth_ports.keys()) - set(cur_ports))

    def is_port_assigned(self, volume_id, port):
        LOG.debug('VOLUME %s: Checking port %s ', volume_id, port)
        cur_ports = self._get_ports(volume_id)
        res = port in cur_ports
        LOG.debug('VOLUME %s: port %s assigned %s',
                  volume_id, port, str(res))
        return res

    def _check_group_mapping(self, vol_id, group_id):
        LOG.debug('Check resource %s in group %s', vol_id, group_id)
        try:
            result, status = self._access_api(tatlin_api.RESOURCE_MAPPING)
        except TatlinAPIException as exp:
            raise exception.VolumeBackendAPIException(
                message=_('Tatlin API exception %s '
                          'while getting resource mapping' % exp.message))

        for entry in result:
            if entry['resource_id'] == vol_id and \
                    entry['host_group_id'] == group_id:
                return True
        return False

    def update_qos(self, vol_id, iops, bandwith):
        pass

    def get_host_id_by_name(self, host_name):
        try:
            result, status = self._access_api(tatlin_api.HOSTS)
            for h in result:
                LOG.debug('For host %s Host name: %s Host ID %s',
                          host_name, h['name'], h['id'])
                if h['name'] == host_name:
                    return h['id']
        except TatlinAPIException as exp:
            message = _('Unable to get host information %s' % exp.message)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        raise exception.VolumeBackendAPIException(
            message='Unable to get host_id for host %s' % host_name)


class TatlinClientV25 (TatlinClientCommon):

    def update_qos(self, vol_id, iops, bandwith):
        path = tatlin_api.RESOURCE % vol_id
        data = {"limit_iops": int(iops),
                "limit_bw": int(bandwith),
                "tags": []}
        try:
            result, status = self._access_api(path, data, 'POST')
            LOG.debug('Responce %s stat %s', result, status)
        except TatlinAPIException as exp:
            message = (_('Unable to update QoS for volume %s due to %s') %
                       (vol_id, exp.message))
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    def export_volume(self, vol_id, eth_ports):
        LOG.debug('VOLUME %s: Export to ports %s started',
                  vol_id, eth_ports)

        to_export = self.get_unassigned_ports(vol_id, eth_ports)
        if not to_export:
            LOG.debug('VOLUME %s: all ports already assigned', vol_id)
            return
        self.export_vol_to_port_list(vol_id, to_export)

        for i in range(self.wait_retry_count):
            if not self.get_unassigned_ports(vol_id, eth_ports):
                LOG.debug('VOLUME %s: Export ports %s finished',
                          vol_id, eth_ports)
                return
            time.sleep(self.wait_interval)

        message = (_('VOLUME %s: Unable to export volume to %s') %
                   (vol_id, eth_ports))
        raise exception.VolumeBackendAPIException(message=message)

    def get_resource_count(self, p_id):
        try:
            result, status = self._access_api(tatlin_api.RESOURCE_COUNT)
        except TatlinAPIException:
            message = _('Unable to get resource count')
            LOG.error(message)
            raise exception.ManageExistingInvalidReference(message)

        poll_resource = 0
        cluster_resources = 0
        for key in result:
            if key == p_id:
                poll_resource = result[key]
            cluster_resources = cluster_resources + result[key]
        return poll_resource, cluster_resources


class TatlinClientV23 (TatlinClientCommon):

    def export_volume(self, vol_id, eth_ports):
        LOG.debug('Export ports %s for volume %s started',
                  eth_ports, vol_id)
        for port in eth_ports:
            LOG.debug('Check port %s for volume %s', port, vol_id)
            if not self.is_port_assigned(vol_id, port):
                try:
                    self.set_port(vol_id, port)
                except TatlinAPIException as e:
                    raise exception.VolumeBackendAPIException(
                        message=e.message)
        LOG.debug('Export ports %s for volume %s finished',
                  eth_ports, vol_id)

        for i in range(self.wait_retry_count):
            if not self.get_unassigned_ports(vol_id, eth_ports):
                LOG.debug('VOLUME %s: Export ports %s finished',
                          vol_id, eth_ports)
                return
            time.sleep(self.wait_interval)

        message = (_('VOLUME %s: Unable to export volume to %s') %
                   (vol_id, eth_ports))
        raise exception.VolumeBackendAPIException(message=message)

    def get_resource_count(self, p_id):
        try:
            response, status = self._access_api(tatlin_api.ALL_RESOURCES)
            if response is not None:
                return 0, len(response)
        except TatlinAPIException:
            message = (_('Unable to get resource list'))
            LOG.error(message)
            return 0, 0
        return 0, 0



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import typing
from typing import Any, Generator, Iterable, Optional, Union  # noqa: H301

from keystoneauth1 import exceptions as ks_exc
from keystoneauth1 import identity
from keystoneauth1 import loading as ka_loading
from keystoneclient import client
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils
import webob
from webob import exc

if typing.TYPE_CHECKING:
    # conditional import to avoid a circular import problem from cinderlib
    from cinder import context
from cinder import exception
from cinder.i18n import _

CONF = cfg.CONF
CONF.import_group('keystone_authtoken',
                  'keystonemiddleware.auth_token.__init__')

LOG = logging.getLogger(__name__)


def _parse_is_public(is_public: Optional[str]) -> Optional[bool]:
    """Parse is_public into something usable.

    * True: List public volume types only
    * False: List private volume types only
    * None: List both public and private volume types
    """

    if is_public is None:
        # preserve default value of showing only public types
        return True
    elif is_none_string(is_public):
        return None
    else:
        try:
            return strutils.bool_from_string(is_public, strict=True)
        except ValueError:
            msg = _('Invalid is_public filter [%s]') % is_public
            raise exc.HTTPBadRequest(explanation=msg)


def is_none_string(val: Any) -> bool:
    """Check if a string represents a None value."""
    if not isinstance(val, str):
        return False

    return val.lower() == 'none'


def remove_invalid_filter_options(
        context: 'context.RequestContext',
        filters: dict,
        allowed_search_options: Iterable[str]) -> None:
    """Remove search options that are not valid for non-admin API/context."""

    if context.is_admin:
        # Allow all options
        return
    # Otherwise, strip out all unknown options
    unknown_options = [opt for opt in filters
                       if opt not in allowed_search_options]
    bad_options = ", ".join(unknown_options)
    LOG.debug("Removing options '%s' from query.", bad_options)
    for opt in unknown_options:
        del filters[opt]


_visible_admin_metadata_keys = ['readonly', 'attached_mode']


def add_visible_admin_metadata(volume) -> None:
    """Add user-visible admin metadata to regular metadata.

    Extracts the admin metadata keys that are to be made visible to
    non-administrators, and adds them to the regular metadata structure for the
    passed-in volume.
    """
    visible_admin_meta = {}

    if volume.get('volume_admin_metadata'):
        if isinstance(volume['volume_admin_metadata'], dict):
            volume_admin_metadata = volume['volume_admin_metadata']
            for key in volume_admin_metadata:
                if key in _visible_admin_metadata_keys:
                    visible_admin_meta[key] = volume_admin_metadata[key]
        else:
            for item in volume['volume_admin_metadata']:
                if item['key'] in _visible_admin_metadata_keys:
                    visible_admin_meta[item['key']] = item['value']
    # avoid circular ref when volume is a Volume instance
    elif (volume.get('admin_metadata') and
            isinstance(volume.get('admin_metadata'), dict)):
        for key in _visible_admin_metadata_keys:
            if key in volume['admin_metadata'].keys():
                visible_admin_meta[key] = volume['admin_metadata'][key]

    if not visible_admin_meta:
        return

    # NOTE(zhiyan): update visible administration metadata to
    # volume metadata, administration metadata will rewrite existing key.
    if volume.get('volume_metadata'):
        orig_meta = list(volume.get('volume_metadata'))
        for item in orig_meta:
            if item['key'] in visible_admin_meta.keys():
                item['value'] = visible_admin_meta.pop(item['key'])
        for key, value in visible_admin_meta.items():
            orig_meta.append({'key': key, 'value': value})
        volume['volume_metadata'] = orig_meta
    # avoid circular ref when vol is a Volume instance
    elif (volume.get('metadata') and
            isinstance(volume.get('metadata'), dict)):
        volume['metadata'].update(visible_admin_meta)
    else:
        volume['metadata'] = visible_admin_meta


def validate_integer(value: int, name: str,
                     min_value: Optional[int] = None,
                     max_value: Optional[int] = None) -> int:
    """Make sure that value is a valid integer, potentially within range.

    :param value: the value of the integer
    :param name: the name of the integer
    :param min_value: the min value of the integer
    :param max_value: the max value of the integer
    :returns: integer
    """
    try:
        value = strutils.validate_integer(value, name, min_value, max_value)
        return value
    except ValueError as e:
        raise webob.exc.HTTPBadRequest(explanation=str(e))


def walk_class_hierarchy(clazz: type,
                         encountered: Optional[list[type]] = None) -> \
        Generator[type, None, None]:
    """Walk class hierarchy, yielding most derived classes first."""
    if not encountered:
        encountered = []
    for subclass in clazz.__subclasses__():
        if subclass not in encountered:
            encountered.append(subclass)
            # drill down to leaves first
            for subsubclass in walk_class_hierarchy(subclass, encountered):
                yield subsubclass
            yield subclass


def _keystone_client(context: 'context.RequestContext',
                     version: tuple[int, int] = (3, 0)) -> client.Client:
    """Creates and returns an instance of a generic keystone client.

    :param context: The request context
    :param version: version of Keystone to request
    :return: keystoneclient.client.Client object
    """
    if context.system_scope is not None:
        auth_plugin = identity.Token(
            auth_url=CONF.keystone_authtoken.auth_url,
            token=context.auth_token,
            system_scope=context.system_scope
        )
    elif context.domain_id is not None:
        auth_plugin = identity.Token(
            auth_url=CONF.keystone_authtoken.auth_url,
            token=context.auth_token,
            domain_id=context.domain_id
        )
    elif context.project_id is not None:
        auth_plugin = identity.Token(
            auth_url=CONF.keystone_authtoken.auth_url,
            token=context.auth_token,
            project_id=context.project_id
        )
    else:
        # We're dealing with an unscoped token from keystone that doesn't
        # carry any authoritative power outside of the user simplify proving
        # they know their own password. This token isn't associated with any
        # authorization target (e.g., system, domain, or project).
        auth_plugin = context.get_auth_plugin()

    client_session = ka_loading.session.Session().load_from_options(
        auth=auth_plugin,
        insecure=CONF.keystone_authtoken.insecure,
        cacert=CONF.keystone_authtoken.cafile,
        key=CONF.keystone_authtoken.keyfile,
        cert=CONF.keystone_authtoken.certfile,
        split_loggers=CONF.service_user.split_loggers)
    return client.Client(auth_url=CONF.keystone_authtoken.auth_url,
                         session=client_session, version=version)


class GenericProjectInfo(object):
    """Abstraction layer for Keystone V2 and V3 project objects"""
    def __init__(self,
                 project_id: str,
                 project_keystone_api_version: str,
                 domain_id: Optional[str] = None,
                 name: Optional[str] = None,
                 description: Optional[str] = None):
        self.id = project_id
        self.keystone_api_version = project_keystone_api_version
        self.domain_id = domain_id
        self.name = name
        self.description = description


def get_project(context: 'context.RequestContext',
                project_id: str) -> GenericProjectInfo:
    """Method to verify project exists in keystone"""
    keystone = _keystone_client(context)
    generic_project = GenericProjectInfo(project_id, keystone.version)
    project = keystone.projects.get(project_id)
    generic_project.domain_id = project.domain_id
    generic_project.name = project.name
    generic_project.description = project.description
    return generic_project


def validate_project_and_authorize(context: 'context.RequestContext',
                                   project_id: str,
                                   policy_check: str,
                                   validate_only: bool = False) -> None:
    target_project: Union[GenericProjectInfo, dict]
    try:
        target_project = get_project(context, project_id)
        if not validate_only:
            target_project = {'project_id': target_project.id}
            context.authorize(policy_check, target=target_project)
    except ks_exc.http.NotFound:
        explanation = _("Project with id %s not found." % project_id)
        raise exc.HTTPNotFound(explanation=explanation)
    except exception.NotAuthorized:
        explanation = _("You are not authorized to perform this "
                        "operation.")
        raise exc.HTTPForbidden(explanation=explanation)
